/*
 * Copyright (C) 2019 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: <Kenneth.Johansson@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdint.h>

#include <linux/jffs2.h>
#include "jffs2-user.h"

#include "log.h"
#include "image.h"
#include "consumer.h"
#include "cmd.h"
#include "crc32.h"

static char device_name[20];
static int mtd_num = -1;
static struct cmd *cmd;

int target_endian = __BYTE_ORDER;

static int sequence;

static int data_start;
static int data_end;
static int data_len;
static int active;


#define SEQUENCE_FILENAME "cferam.000"

static char *cbuf;     /* temp buffer used to enough data to serch for direntry */
static int   cb_len;   /* total len of cbuf = direntry*2 */
static int   cb_vlen;  /* len of valid data */
static int   cb_slen;  /* len of struct jffs2_raw_dirent + filename */

#if 0
static const char *cmd_name={"dd"};
static const char * const cmd_argv[]={"dd", "of=vmlinux_jffs2.img", "status=none", NULL};
#else
static const char *cmd_name={"imagewrite"};
static const char *cmd_argv[]={"imagewrite","place holder for device name" ,"-c", "-i", NULL};
#endif

#if 0
static void dump_cbuf( void );
static void dump_cbuf( void )
{
        int len;

        printf(" cbuf [");
        for ( len = 0; len < cb_vlen ; len++) {
                printf(" 0x%02x ", (unsigned char) cbuf[len] );
        }
        printf("]");
}
#endif

static int init( struct img_header *hdr)
{

        if (! active)
                return 0;

        /* sanity check. sequence number has to be given */
        if (!sequence){
                fprintf(stderr, "brcm kernel needs the sequence number to use [-s|--sequence]\n");
                exit(EXIT_FAILURE);
        }

        /* read out header data */
        data_start = get_int_val(hdr, "vmlinuxofs");
        data_len   = get_int_val(hdr, "vmlinux");

        if (data_start == -1 || data_len == -1 ){
                fprintf(stderr, "Could not find linux kernel data in image\n");
                fprintf(stderr, "Aborting\n");
                exit(EXIT_FAILURE);
        }

        data_end = data_start + data_len;

        /* cbuf is large enough to contain two directory entries with SEQUENCE_FILENAME */
        cb_slen = sizeof(struct jffs2_raw_dirent) + strlen(SEQUENCE_FILENAME);
        cb_len = cb_slen * 2;
        cbuf = malloc( cb_len );

        if (!cbuf) {
                fprintf(stderr, "brcm kernel cound not allocate search buffer\n");
                exit(EXIT_FAILURE);
        }
        /* fill in device name */
        snprintf(device_name, 20, "/dev/mtd%d", mtd_num);
        cmd_argv[1] = device_name;

        cmd = fork_cmd(cmd_name, cmd_argv);

        return 0;
}

static void usage(void)
{
        printf("      -k, --kernel=DEV\tProgram the linux kernel to MTD device number\n");
        printf("      -s, --sequence=NUM\tSequence number to add to cferam.000\n");
}

static void kernel_opt( int argc, char **argv){
        int ch;
        char *endptr;

        while (1) {
                int option_index = 0;
                static struct option long_options[] = {
                        {"kernel",  required_argument, 0, 'k'},
                        {"sequence",required_argument, 0, 's'},
                        {0, 0, 0, 0}
                };

                ch = getopt_long(argc, argv, "k:s:",
                                 long_options, &option_index);

                if (ch == -1)
                        break;

                switch (ch) {
                case 'k':
                        mtd_num = strtol(optarg, &endptr, 10);
                        if ( optarg == endptr ) {
                                fprintf(stderr, "Wrong format for mtd partition number.Was [%s] "
                                        "should be X where X is a decimal number\n", optarg);
                                exit (EXIT_FAILURE);
                        }
                        active = 1;
                        break;
                case 's':
                        sequence = strtol(optarg, 0, 0);
                        break;
                default:
                        ;
                }
        }
}
static int scan_sequence( struct range_data *rd, char *buf);

static int match(char *buf,int times)
{
        int i;

        for (i=0 ; i<times; i+=1) {
                struct jffs2_raw_dirent *pdir= (struct jffs2_raw_dirent *)( buf + i );

                /* find dir entry for SEQUENCE_FILENAME file. */
                if( je16_to_cpu(pdir->magic) == JFFS2_MAGIC_BITMASK ) {
                        int name_len= strlen(SEQUENCE_FILENAME);
                        if( je16_to_cpu(pdir->nodetype) == JFFS2_NODETYPE_DIRENT) {
                                if(pdir->nsize == name_len) {
                                        if (! strncmp((char*)pdir->name, SEQUENCE_FILENAME, name_len - 3 )) { /* do not compare last 3 chars of name */
						if( je32_to_cpu(pdir->ino) != 0 ) {
                                                        /* we got match */
#if 0
                                                        printf("name    : %.10s\n",pdir->name);/* string is not terminated with 0, so garbage here */
                                                        printf("name_crc: %x\n",je32_to_cpu(pdir->name_crc));
                                                        printf("nsize   : %d\n",pdir->nsize);
                                                        printf("totlen  : %d\n",je32_to_cpu(pdir->totlen));
#endif
                                                        return i;
                                                }
                                        }
                                }
                        }
                }
        }
        return -1;
}

static void update_direntry_inplace(char * buf){
        struct jffs2_raw_dirent *pdir = (struct jffs2_raw_dirent *)buf;
        char tmp[20];

        /* dirent has no terminating 0 on strings so handle that using temp buffer */
        snprintf( tmp, strlen(SEQUENCE_FILENAME) + 1, "cferam.%03d", sequence);
        memcpy( (char*)pdir->name, tmp, strlen(SEQUENCE_FILENAME) );

        /* recalculate name crc */
	pdir->name_crc = cpu_to_je32(crc32(0, pdir->name, strlen(SEQUENCE_FILENAME)));
}

static int found;

static int scan_sequence( struct range_data *rd, char *buf)
{
        int match_idx;

        while (1) {

                int num;

                /* Try to fill entire cbuf with valid data */
                int need = cb_len - cb_vlen;
                if (need > rd->src_len )
                        need = rd->src_len;

                memcpy(cbuf + cb_vlen, buf + rd->src_offset, need);

                /* update src buffer after move of data to cbuf */
                rd->src_offset += need;
                rd->src_len -= need;
                cb_vlen += need;


                /* enough data to do a match ? */
                if (cb_vlen  < cb_slen) {
                        return 0; /* nope */
                }

                /* how many matches can be done inside cbuf */
                num = cb_vlen - cb_slen + 1;

                /* match */
                if ( -1 != (match_idx = match(cbuf, num)) ) {
                        update_direntry_inplace(cbuf + match_idx);

                        /* write out all data */
                        write(cmd->io_out, cbuf, cb_vlen);
                        cb_vlen = 0;

                        return 1;
                }

                /* no match so far.*/
                /* write out already scanned data and move not scanned data to front of buffer. */
                if (num ){
                        /* now we can write the previous data in cbuf as we
                           know it contain no part that needs to change */
                        write(cmd->io_out, cbuf, num);

                        /* remove already written data from cbuf */
                        cb_vlen -= num;
                        memmove(cbuf, cbuf + num, cb_vlen);
                }
        }
}

static int write_data(char *buf,int offset, int len)
{
        struct range_data rd;

        if (!active)
                return 0;

        rd = adjust_range(data_start, data_end, offset, len);
        if (rd.src_len == 0 )
                return 0;

        /* scan for direntry */
        if (!found)
                found = scan_sequence(&rd,buf);

        /* is there still some data left to write */
        if (rd.src_len)
                write(cmd->io_out, buf + rd.src_offset, rd.src_len);


        return 0;
}

static struct ops my_ops = {
        .description = "program brcm linux kernel to MTD partition",
        .init = init,
        .usage = usage,
        .opt = kernel_opt,
        .data = write_data,

};

CONSUMER(my_ops);
