/*
 * Copyright (C) 2019 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: <Kenneth.Johansson@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
/* cfe is split into two partitions. first one is 128Kb and the next is used for the rest. */

/* cfe writing can only be run in post. this is so the digital signing and md5sum can      */
/* be checked before any code is actually written to flash                                 */

#include <stdio.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>

#include "log.h"
#include "image.h"
#include "consumer.h"
#include "cmd.h"


static int mtd_first = -1;
static int mtd_second = -1;

static int data_start;
static int data_end;
static int data_len;
static int active;
static int burn;

static const char *cmd_dump_prg={"dd"};
static const char *cmd_dump_argv[]={"dd", NULL, NULL};
static char cmd_dump_name[100];
static struct cmd *cmd_dump;

static char *filename;

static char *cfe_data;
static int   cfe_data_len;

static int major = -1;
static int minor = -1;
static int do_version;

/* used to find where nvram starts in CFE image */
static int nvram_offset;

static int init( struct img_header *hdr)
{
        if (! active)
                return 0;

        if (do_version){
                char *next;
                char *endptr;
                char *s;
                int new_major;
                int new_minor;
                s = get_str_val(hdr, "cfever");

                new_major = strtol(s, &next, 10);
                if ( s == next ) {
                        fprintf(stderr, "Wrong format for cfe version number in image header.Was [%s] "
                                "should be X.X where X is a decimal number\n", s);
                        exit (EXIT_FAILURE);
                }
                new_minor = strtol(next+1, &endptr, 0);
                if (next+1 == endptr){
                        fprintf(stderr, "Wrong format for cfe version number in image header.Was [%s] "
                                "should be X.X where X is a decimal number\n", s);
                        exit (EXIT_FAILURE);
                }
                /* if version is too low deactivate the cfe programming */
                if (new_major <= major && new_minor <= minor ) {
                        printf("New cfe version %d.%d not newer than current %d.%d, Skipping update of CFE\n",
                               new_major,new_minor,major,minor);
                        active=0;
                        burn=0;
                        return 0;
                }

        }
        /* read out header data */
        data_start = get_int_val(hdr, "cfeofs");
        data_len   = get_int_val(hdr, "cfe");

        if (data_start == -1 || data_len == -1 ){
                fprintf(stderr, "cfe extractor could not get start or length for cfe section\n");
                fprintf(stderr, "Aborting\n");
                exit(EXIT_FAILURE);
        }

        data_end = data_start + data_len;

        if (mtd_first != -1 ) {

                /* Alloc data buffer */
                cfe_data = malloc(data_len);
                if (! cfe_data) {
                        fprintf(stderr, "Can not allocate temporary buffer for CFE image data\n");
                        fprintf(stderr, "Aborting\n");
                        exit(EXIT_FAILURE);
                }
        }

        if (filename) { /* dump to filename */
                snprintf(cmd_dump_name, 100, "of=%s", filename);
                cmd_dump_argv[1] = cmd_dump_name;
                cmd_dump = fork_cmd(cmd_dump_prg, cmd_dump_argv);
        }

        return 0;
}

static void usage(void)
{
        printf("      -c, --cfe=num,num\tprogram CFE to MTD devices num,num.\n");
        printf("                       \tOnly the mtd partition number should be given.\n");
        printf("      -C, --cfe_dump=NAME\tExtract the CFE image to file\n");
        printf("      -m, --cfe_ver=num,num\tNew CFE version needs to be newer than this MAJOR,MINOR\n");
        printf("                           \tElse it wll not be programmed. Omit this option to write any version\n");
        printf("      -N, --nvram=offset\tOffset in cfe where the nvram is stored.\n");

}

static void usage_err(int ch)
{
        fprintf(stderr, "Option [-%c/",ch);
        switch(ch) {
        case 'c':
                fprintf(stderr, "cfe burn");
                break;
        case 'C':
                fprintf(stderr, "cfe dump");
                break;
        case 'N':
                fprintf(stderr, "nvram offset");
                break;
        }
        fprintf(stderr, "] require argument\n");
        exit(EXIT_FAILURE);
}

static void cfe_opt( int argc, char **argv)
{
        int ch;
        char *two=0;

        while (1) {
                int option_index = 0;
                static struct option long_options[] = {
                        {"cfe", required_argument, 0, 'c'},
                        {"cfe_dump", required_argument, 0, 'C'},
                        {"cfe_ver", required_argument, 0, 'm'},
                        {"nvram", required_argument, 0, 'N'},
                        {0, 0, 0, 0}
                };

                ch = getopt_long(argc, argv, "c:C:N:m:",
                                 long_options, &option_index);
                if (ch == -1)
                        break;

                switch (ch) {
                case 'c':
                        mtd_first = strtol(optarg, &two, 10);
                        if ( optarg == two ) {
                                fprintf(stderr, "Wrong format for mtd partition number.Was [%s] "
                                        "should be X,X where X is a decimal number\n", optarg);
                                exit (EXIT_FAILURE);
                        }
                        if (two > optarg) {
                                char *endptr=0;
                                mtd_second = strtol(two+1, &endptr, 0);
                                if (two+1 == endptr){
                                        fprintf(stderr, "Wrong format for mtd partition number.Was [%s] "
                                                "should be X,X where X is a decimal number\n", optarg);
                                        exit (EXIT_FAILURE);
                                }
                        }

                        if (mtd_first >= 0 ){
                                burn = 1;
                                active = 1;
                        }

                        break;
                case 'm':
                        major = strtol(optarg, &two, 10);
                        if ( optarg == two ) {
                                fprintf(stderr, "Wrong format for cfe version number.Was [%s] "
                                        "should be X,X where X is a decimal number\n", optarg);
                                exit (EXIT_FAILURE);
                        }
                        if (two > optarg) {
                                char *endptr=0;
                                minor = strtol(two+1, &endptr, 0);
                                if (two+1 == endptr){
                                        fprintf(stderr, "Wrong format for cfe version number.Was [%s] "
                                                "should be X,X where X is a decimal number\n", optarg);
                                        exit (EXIT_FAILURE);
                                }
                        }
                        if (major >= 0 ){
                                do_version=1;
                        }
                        break;
                case 'C':
                        filename = optarg;
                        active = 1;
                        break;
                case 'N':
                        nvram_offset = strtol(optarg,0,0);
                        break;
                case '?':
                        switch(optopt) {
                        case 'c':
                        case 'C':
                        case 'N':
                                usage_err(optopt);
                                break;
                        }
                default:
                        ;
                }
        }


        /* sanity check. we need the nvram offset otherwise it wont work. */
        if (burn && nvram_offset == 0){
                fprintf(stderr, "Burning CFE needs the nvram offset\n");
                fprintf(stderr, "ARM = -N 66944, MIPS = -N 1408\n");
                exit (EXIT_FAILURE);
        }
}

static int write_data(char *buf,int offset, int len)
{
        struct range_data rd;

        if (!active)
                return 0;

        rd = adjust_range(data_start, data_end, offset, len);

        if (filename ) {
                if (rd.src_len)
                        write(cmd_dump->io_out, buf + rd.src_offset, rd.src_len);
        }
        if (mtd_first != -1) {
                if (rd.src_len) {
                        memcpy(cfe_data + cfe_data_len, buf + rd.src_offset, rd.src_len);
                        cfe_data_len += rd.src_len;
                }
        }

        return 0;
}

/* cfe has a section in the image that is used as nvram that the factory programs */
/* while inteno is using a copy if that it is still important to copy the factory default */
/* in case of a full erase */
static void copy_nvram( void )
{
        int fd, res;
        char cfe_dev[100];
        off_t off;

        if (!burn)
                return ;

        snprintf(cfe_dev, 100, "/dev/mtd%d",mtd_first);
        fd = open(cfe_dev, O_RDONLY, 0);
        if (fd == -1 ) {
                fprintf(stderr, "Could not open %s for reading out default nvram data\n", cfe_dev);
                exit (EXIT_FAILURE);
        }
        off = lseek(fd, nvram_offset, SEEK_SET);
        if (off == -1) {
                fprintf(stderr, "Could not seek to %d on cfe partition %s\n", nvram_offset, cfe_dev);
                exit (EXIT_FAILURE);
        }

        /* store the cfe data from MTD into the in ram copy of new CFE */
        res = read(fd, cfe_data + nvram_offset, 1024);
        if (res == -1 ) {
                fprintf(stderr, "Could not read default NVRAM out od CFE\n");
                exit (EXIT_FAILURE);
        }
}

#if 0
static void imagewrite_result(struct cmd *c, char *cfe_dev)
{
        char result[1024*4];

        read(c->io_in, result, sizeof(result));
        printf("result from imagewrite to %s \n-----------------------\n", cfe_dev);
        printf("%s",result);
        printf("\n-----------------------\n");
}
#else
#define imagewrite_result(x,y)
#endif

static void burn_image( void )
{
        char cfe_dev[100];
        int len_to_write;
        int status;

        const char *cmd_name={"imagewrite"};
        const char *cmd_argv[]={"imagewrite", "place holder for device name" , "-i", NULL};
        struct cmd *cmd;

        if (!burn)
                return;

        snprintf(cfe_dev, 100, "/dev/mtd%d", mtd_first);

        cmd_argv[1] = cfe_dev;
        cmd = fork_cmd(cmd_name, cmd_argv);

        /* Write out to first mtd. max 128kb */
        if (data_len > 1024*128)
                len_to_write = 1024*128;
        else
                len_to_write = data_len;

        write(cmd->io_out, cfe_data, len_to_write);
        close(cmd->io_out);
        waitpid(cmd->cpid, &status, 0);

        imagewrite_result(cmd, cfe_dev);

        /* Are we done. is there more data to write ? */
        /* iopsys CFE is split up into two different blocks */
        if ((data_len - len_to_write) > 0) {
                snprintf(cfe_dev, 100, "/dev/mtd%d", mtd_second);
                cmd_argv[1] = cfe_dev;
                cmd = fork_cmd(cmd_name, cmd_argv);

                len_to_write = data_len - 1024*128;
                write(cmd->io_out, cfe_data + 1024*128, len_to_write);
                close(cmd->io_out);
                waitpid(cmd->cpid, &status, 0);

                imagewrite_result(cmd, cfe_dev);
        }
}

static struct ops my_ops = {
        .description = "Extract CFE image",
        .init = init,
        .usage = usage,
        .opt = cfe_opt,
        .data = write_data,
        .finish = copy_nvram,
        .post = burn_image,
};

CONSUMER(my_ops);
