/*
 * Copyright (C) 2019 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: <Kenneth.Johansson@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "image.h"


static int active;

/* every block read from the image is put in a buffer.
   This function returns how much of the current buffer is of intereset.
   input is start and end offset of the entire image that is of interest.
   offset is the current buffers offset and len is the current buffer len.
*/
struct range_data adjust_range(int start,int end, int offset, int len)
{
        struct range_data rd;

        rd.src_len = 0;

        rd.src_offset = start - offset;
        if (rd.src_offset < 0)
                rd.src_offset = 0;

        if (rd.src_offset > len) {
                return rd; /* invalid range, no data */
        } else {
                rd.src_len = len - rd.src_offset;
                if (offset + len > end)
                        rd.src_len = rd.src_len - (offset + len - end);
                if (rd.src_len < 0)
                        rd.src_len = 0;
                return rd; /* valid range*/
        }

}


/* this allocate a new string, remember to free */
char *get_str_val(struct img_header *hdr, const char *key)
{
        char *s;
        char *e;

        /* find second line */
        s = strchr(hdr->raw_buf, '\n');
        if (!s) return NULL;
        s++;

        /* find key at start of line */
        while(1) {
                if (! strncmp(s, key, strlen(key)) && s[strlen(key)] == ' ') {
                        break;
                }else {
                        s = strchr(s, '\n');
                        if (!s) return NULL;
                        s++;
                }
        }

        /* Key string found. Now extract the value string */
        s += strlen(key) + 1;
        e = strchr(s, '\n');

        return strndup(s, e-s);
}

/* -1 is error code */
int get_int_val(struct img_header *hdr, const char *key)
{
        char *p;
        int num;

        p = get_str_val(hdr, key);
        if (!p ) return -1;

        num = strtol(p, NULL, 10);

        free(p);
        return num;
}

static void parse_headers(struct img_header *hdr)
{
        /* y2 has IntenoBlob */
        if (! strncmp(hdr->raw_buf,"IntenoBlob",strlen("IntenoBlob"))) {
                hdr->type = IMG_Y2;
        /* y3 has IntenoIopY */
        } else if (! strncmp(hdr->raw_buf,"IntenoIopY",strlen("IntenoIopY"))){
                hdr->type = IMG_Y2;
        }else {
                fprintf(stderr, "Image format is not recognized \nAborting!\n");
                exit(1);
        }

}

void img_print_hdr(struct img_header *hdr)
{
        if (hdr->raw_buf)
                puts(hdr->raw_buf);
}

/* Return a struct to header data */
struct img_header *img_init(int input)
{
        struct img_header *hdr;
        int bytes=0;

        hdr = malloc(sizeof(struct img_header));
        if (!hdr) goto mem_fail;
        memset(hdr, 0, sizeof(struct img_header));

        hdr->raw_buf = malloc(HEADERS_SIZE + 1);
        if (!hdr->raw_buf) goto mem_fail;

        while (bytes < HEADERS_SIZE) {
                int len;
                len = read(input, hdr->raw_buf + bytes , HEADERS_SIZE - bytes );
                if (len == -1 || len == 0 ) {
                        fprintf(stderr, "Tried to read %d bytes of header data from image file but only got %d. ",
                                HEADERS_SIZE, bytes);
                        if (len == -1 )
                                perror("");
                        else
                                fprintf(stderr, "\n");
                        fprintf(stderr, "Aborting! \n");
                        exit(1);
                }
                bytes += len;
        }
        /* Use the extra byte allocated for raw buf as a str termination point */
        /* so we know any str search is done only on the header data.          */
        hdr->raw_buf[HEADERS_SIZE] = 0;

        parse_headers(hdr);

        active = 1;

        return hdr;
mem_fail:
        fprintf(stderr, "Memory alloc fail. Now way to continue\n");
        exit(1);
}
