/*
 * Copyright (C) 2019 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: <Kenneth.Johansson@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#include <config.h>
#include <stdio.h>
#include <getopt.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/wait.h>
#include <string.h>

#include "config.h"
#include "log.h"
#include "cmd.h"
#include "image.h"
#include "consumer.h"

int debug_level;
int fd_image;
int verbose;

static void print_usage(const char *prg_name);
void cpy_argv(int argc, char *argv[]);

static void print_usage(const char *prg_name) {

        struct ops *ops;

        printf("Usage: %s [OPTION]..\n", prg_name);
        printf("  Options: \n");
        printf("      -f, --file=NAME\timage file to use. If not specified stdin is used.\n");

        /* print out consumer args */
        list_for_each_entry(ops, &consumers, list)
                if (ops->usage) ops->usage();

        printf("      -p, --print\tPrint headers.\n");
        printf("      -v\t\tVerbose output.\n");
        printf("      -d, --debug=NUM\tSet debug level. Higher = more output\n");
        printf("      -h\t\tShow this help screen.\n");
        printf("\n");
}

/* getopt rearanges the array so we need a copy to work on */
char **argv_cpy;
void cpy_argv(int argc, char *argv[]) {

        static int allocated=0;
        int len = sizeof(void*) * (argc + 1);

        if (!allocated) {
                argv_cpy = malloc( len );
                allocated = 1;
        }

        memcpy(argv_cpy , argv, len);
}

#if 0
static dump_argv(int argc, char *argv[])
{
        int i;

        for (i=0; i<argc;i++){
                fprintf(stderr, "%d %s\n",i,argv[i]);
        }
}
#endif

int main(int argc, char *argv[])
{
        int ch;
        int data_len;
        int data_read;
        int print = 0;
        char *filename = NULL;
        struct img_header *hdr;
        struct ops *ops;

        cpy_argv(argc, argv);

/************************************************************************/
/*  Stage 0. Handle command line parameters                             */
/************************************************************************/
        opterr = 0;
        while (1) {
                int option_index = 0;
                static struct option long_options[] = {
                        {"debug", required_argument, 0, 'd'},
                        {"file",  required_argument, 0, 'f'},
                        {"print",       no_argument, 0, 'p'},
                        {"verbose",     no_argument, 0, 'v'},
                        {"help",        no_argument, 0, 'h'},
                        {0, 0, 0, 0}
                };

                ch = getopt_long(argc, argv_cpy, "d:f:pvh",
                                 long_options, &option_index);
               if (ch == -1)
                   break;

               switch (ch) {
               case 'd':
                        debug_level = strtol(optarg, 0, 0);
                        break;
               case 'f':
                       filename = optarg;
                       break;
               case 'p':
                       print = 1;
                       break;
               case '?':
                       break;
               case 'h':
               default:
                       print_usage(argv[0]);
                       exit(0);
                }
        }


        /* handle parsing for all consumers */
        list_for_each_entry(ops, &consumers, list)
                if (ops->opt) {
                        optind = 1;
                        cpy_argv(argc, argv);
                        ops->opt(argc, argv_cpy);
                }

        /* debug: print all consumer modules we have */
        if (debug_level > 0 ){
                printf("------ Available modules\n");
                list_for_each_entry(ops, &consumers, list)
                        if (ops->description)
                                printf("%s\n",ops->description);
                printf("------------------------\n");
        }

/************************************************************************/
/*  Stage 1. read header from image file.                               */
/************************************************************************/

        /* are we reading from a file or stdin */
        if (filename) {
                fd_image = open(filename, O_RDONLY);
                if (fd_image == -1){
                        fprintf(stderr, "Tried to open file [%s]. ",filename);
                        perror("");
                        exit(EXIT_FAILURE);
                }
        } else {
                fd_image=STDIN_FILENO;
        }

        /* read in headers from image to find out what we have */
        /* this only returns valid data */
        hdr = img_init(fd_image);

        if ( print ){
                img_print_hdr(hdr);
                exit(EXIT_SUCCESS);
        }

        /* Init all sub modules once we have headers. */
        list_for_each_entry(ops, &consumers, list)
                if (ops->init) ops->init(hdr);

/************************************************************************/
/*  Stage 2. read image file data and pipe to sub commands              */
/************************************************************************/

        /* header is already read in so pipe that data first */

        list_for_each_entry(ops, &consumers, list)
                if (ops->data) ops->data(hdr->raw_buf, 0 ,1024);

        data_read = 1024;

        /*  now loop over the rest of the file */

        data_len = get_int_val(hdr, "size");
        if (-1 == data_len ) {
                fprintf(stderr, "Could not read total image size. Aborting\n");
                exit(EXIT_FAILURE);
        }


/* set BLOCKSIZE 1 to test that nothing assume data is read from a file */
//#define BLOCK_SIZE 1024
#define BLOCK_SIZE 1024

        while ( data_read < data_len ) {
                int len;
                char buf[BLOCK_SIZE];

                /* clamp data to 1 KB blocks */
                len = data_len - data_read;
                len > BLOCK_SIZE ? len=BLOCK_SIZE : len;

                /* read from image */
                len = read(fd_image, buf, len);
                if (len < 0 ) {
                        fprintf(stderr, "Reading mage data error");
                        perror("");
                        exit(EXIT_FAILURE);
                }

                /* handle case where header indicate more data than in file */
                if (len == 0) {
                        fprintf(stderr, "Reading image data error: got %d bytes wanted %d\n", data_read, data_len );
                        perror("");
                        exit(EXIT_FAILURE);
                }

                if (debug_level > 5)
                        printf("Write buff size = %d\n",len);
                /* pipe to sub commands */
                list_for_each_entry(ops, &consumers, list)
                        if (ops->data) ops->data( buf, data_read, len);

                data_read += len;
        }

/************************************************************************/
/*  Stage 3. Finish commands and get results                            */
/************************************************************************/

        list_for_each_entry(ops, &consumers, list)
                if (ops->finish) ops->finish();


/************************************************************************/
/*  Stage 4. post. Things that must run after integrity checks          */
/************************************************************************/

/*here we program the boot loader and run the config extraction*/
        list_for_each_entry(ops, &consumers, list)
                if (ops->post) ops->post();

        return 0;
}
