/*
 * Copyright (C) 2019 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: <Kenneth.Johansson@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "image.h"
#include "consumer.h"
#include "cmd.h"

static struct cmd *cmd;

static int active;
static int data_len;
static char image_checksum[32 + 1];

static const char *cmd_name={"md5sum"};
static const char * const cmd_argv[]={"md5sum", NULL};

static void store_csum(char *buf,int offset, int len)
{
        int dst_offset;
        int src_offset;
        int cpy_len;

        /* buf is only data from the last 32 bytes */
        if ( offset >= data_len ){
                dst_offset = offset - data_len;
                cpy_len = len;
                memcpy(image_checksum + dst_offset, buf, cpy_len);
                return;
        }

        /* offset < data_len */
        /* buf is containing data + parts|all of the last 32 bytes */
        src_offset = data_len - offset;
        cpy_len = len - src_offset;
        memcpy(image_checksum, buf + src_offset, cpy_len);

}

static int init(struct img_header *hdr)
{
        char *p;

        /* integrity */
        p = get_str_val(hdr, "integrity");
        if (p) {
                if (! strcmp("MD5SUM", p)) {
                        active = 1;
                }
        }else {
                return 0;
        }

        /* md5sum is over the whole file minus 32 bytes at the end */
        data_len = get_int_val(hdr, "size");
        if (-1 == data_len ){
                active = 0;
                fprintf(stderr, "md5sum: Could not read total image size. Aborting\n");
                exit(EXIT_FAILURE);
        }else
                data_len -= 32;

        cmd = fork_cmd(cmd_name, cmd_argv);
        return 0;
}

static int write_data(char *buf,int offset, int len)
{
        if (!active)
                return 0;

        /* make sure we do not include last 32 bytes in caclulation but also store it to compare against .*/
        if (offset >= data_len){
                store_csum(buf, offset, len);
                return 0;
        }

        if ( offset + len >= data_len){
                store_csum(buf, offset, len);
                len = data_len - offset;
        }

        /* pipe data to md5sum child process */
        while (1){
                int res;
                res = write(cmd->io_out, buf, len);
                if (res < 0 ) {
                        /* if md5sum fails there is no point in continuing. Image can not be used anyway */
                        fprintf(stderr, "Writing to md5sum");
                        perror("");
                        exit(EXIT_FAILURE);
                }
                if (res != len )
                        len = len - res;
                else
                        break; /* whole buffer is written */
        }
        return 0;
}

static void finish(void)
{
        char buf[32 + 1];

        buf[32] = 0;

        close(cmd->io_out);
        read(cmd->io_in, buf, 32);

//        printf("md5sum calculated [%s]\n", buf);
//        printf("md5sum from file  [%s]\n", image_checksum);

        /* if csum not ok quit*/
        if (strcmp(buf,image_checksum)) {
                fprintf(stderr, "md5sum: Not matching got %s wanted %s. Aborting\n", buf, image_checksum);
                exit(EXIT_FAILURE);
        }
}

static struct ops my_ops = {
        .description = "MD5 checksum verification",
        .init = init,
        .data = write_data,
        .finish = finish,
};

CONSUMER(my_ops);
