/*
 * Copyright (C) 2019 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: <Kenneth.Johansson@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include "log.h"
#include "image.h"
#include "consumer.h"
#include "cmd.h"

#include "crc32.h"
#include "ubi-media.h"
#include "mtd_swab.h"

#ifdef DEBUG
# undef DEBUG
# define DEBUG(args)	do { printf args; } while (0)
#else
# define DEBUG(args)
#endif


static int peb_size = 128*1024; /* physical block size. We only have boards with 128kB for now so hardcode it */
static char *peb;		/* pointer to one block */
static int peb_len;		/* total length of valid data in peb */


static int data_start;
static int data_end;
static int data_len;

static int active;
static int dump;

static struct cmd *cmd;
static char *filename;

static const char *dump_cmd_prg={"dd"};
static const char *dump_cmd_argv[]={"dd", NULL, NULL};
static char dump_cmd_name[100];

static const char *cmd_name={"ubiupdatevol"};
static const char *cmd_argv[]={"ubiupdatevol","", "", "-",NULL};

static char ubifs_size[100];
static char ubifs_device[100];

static int no_ubi;	/* set this if there is no ubi image just an ubifs */

static int init( struct img_header *hdr)
{

        if (! active)
                return 0;


        /* read out ubi header data */
        data_start = get_int_val(hdr, "ubiofs");
        data_len   = get_int_val(hdr, "ubi");

	/* if no ubi maybe an ubifs */
	if (data_len == -1 || data_len ==0 ) {
		no_ubi = 1;
		data_start = get_int_val(hdr, "ubifsofs");
		data_len   = get_int_val(hdr, "ubifs");
	}

        if (data_start == -1 || data_len == -1 || data_len == 0 ){
                fprintf(stderr, "Could not find ubi/ubifs data in image\n");
                fprintf(stderr, "Aborting\n");
                exit(EXIT_FAILURE);
        }
	data_end = data_start + data_len;

	if (dump) {
		snprintf(dump_cmd_name, 100, "of=%s", filename);
                dump_cmd_argv[1] = dump_cmd_name;

		cmd = fork_cmd(dump_cmd_prg, dump_cmd_argv);
	}else {
		if (no_ubi) {
			snprintf(ubifs_size,100,"-s %d", data_len);
		}else{
			/* magic: we need to gitve ubiupdatevol a size for the ubifs image */
			/* we only have the size of the UBI image and we can not scan the data */
			/* this caclulation only works for images crated from iopsys build not on a real ubifs image */
			int pebs = data_len/peb_size;
			int size = (pebs-2)*(peb_size-4096);
			snprintf(ubifs_size,100,"-s %d", size);
		}
		cmd_argv[1] = ubifs_device;
		cmd_argv[2] = ubifs_size;

		cmd = fork_cmd(cmd_name, cmd_argv);
	}

	peb = malloc(peb_size);
	if ( NULL == peb ){
                fprintf(stderr, "Could not allocate peb buffer\n");
                fprintf(stderr, "Aborting\n");
                exit(EXIT_FAILURE);
	}

        return 0;
}

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

struct ubi_info {
	off_t size;
	int vid_hdr_offset;
	int data_offset;
	int vol_id;
	int lnum;		/* logical eraseblock number */
};

static const char *vol_name = "rootfs_0";
static	int vol_ix;

/* in every PEB block there is an UBI header. parse it. */
static int read_ubi_hdr(struct ubi_info *ubi_info)
{
	struct ubi_ec_hdr *ec_hdr;
	struct ubi_vid_hdr *vid_hdr;
	unsigned int magic, crc;
	struct ubi_vtbl_record *vtbl_rec;

	/* read out and verify UBI erase counter header */
	ec_hdr = (struct ubi_ec_hdr *)peb;
	magic = be32_to_cpu(ec_hdr->magic);
	if (magic != UBI_EC_HDR_MAGIC) {
		fprintf(stderr, "Wrong MAGIC in UBI image. Can not continue\n");
		exit(EXIT_FAILURE);
	}

	crc = crc32(UBI_CRC32_INIT, ec_hdr, UBI_EC_HDR_SIZE_CRC);
	if (be32_to_cpu(ec_hdr->hdr_crc) != crc) {
		fprintf(stderr, "Wrong CRC in UBI image. Can not continue\n");
		exit(EXIT_FAILURE);
	}

	ubi_info->vid_hdr_offset = be32_to_cpu(ec_hdr->vid_hdr_offset);
	ubi_info->data_offset = be32_to_cpu(ec_hdr->data_offset);

	if (debug_level > 0 )
		printf("UBI: vid_hdr_offset=%u, data_offset=%u\n", ubi_info->vid_hdr_offset, ubi_info->data_offset);

	/* read out and verify UBI volume identifier header */
	vid_hdr = (struct ubi_vid_hdr *)(peb + ubi_info->vid_hdr_offset);

	magic = be32_to_cpu(vid_hdr->magic);
	if (magic == 0xfffffffful) {
		fprintf(stderr, "UBI was empty. Can not continue\n");
		exit(EXIT_FAILURE);
	} else if (magic != UBI_VID_HDR_MAGIC){
		fprintf(stderr, "UBI bad MAGIC. Can not continue\n");
		exit(EXIT_FAILURE);
	}

	crc = crc32(UBI_CRC32_INIT, vid_hdr, UBI_VID_HDR_SIZE_CRC);
	if (be32_to_cpu(vid_hdr->hdr_crc) != crc) {
		fprintf(stderr, "Wrong CRC in UBI vid hdr. Can not continue\n");
		exit(EXIT_FAILURE);
	}

 	ubi_info->vol_id = be32_to_cpu(vid_hdr->vol_id);
	ubi_info->lnum = be32_to_cpu(vid_hdr->lnum);

	if (debug_level > 0 )
		printf("vol_id=0x%08x, lnum=%u\n", ubi_info->vol_id, ubi_info->lnum);


	/* if this is the layout volume find out what volume id we should use for rootfs volume */
	if ( ubi_info->vol_id == UBI_LAYOUT_VOLUME_ID) {

		/* find volume id for "rootfs_0" */
		/* BUG: this code assumes only one volume exist, so volume id is always 0 */
		/* not future proof but works now  */
		if ( peb_len < ubi_info->data_offset + (int)sizeof(struct ubi_vtbl_record) )
			return 0; /*no*/

		vtbl_rec = (struct ubi_vtbl_record*) (peb + ubi_info->data_offset);
		crc = crc32(UBI_CRC32_INIT, vtbl_rec, UBI_VTBL_RECORD_SIZE_CRC);

		if (be32_to_cpu(vtbl_rec->crc) != crc) {
			fprintf(stderr, "Wrong CRC for ubi volume table. aborting\n");
			exit(EXIT_FAILURE);
		}

		if (strncmp(vol_name, (char *)vtbl_rec->name, UBI_VOL_NAME_MAX)) {
			fprintf(stderr, "Wrong volume name [%s] wanted [%s]. aborting\n",(char *)vtbl_rec->name, vol_name);
			exit(EXIT_FAILURE);
		}
		vol_ix = 0;
	}

	return 1;
}

/* Read in one erase block size buffer at a time and handle that */
/* volume name is rootfs*/
static void extract_ubifs(void)
{
	struct ubi_info ubi_info;

	if (read_ubi_hdr(&ubi_info)) {
		if (ubi_info.vol_id == vol_ix ){
			write(cmd->io_out, peb + ubi_info.data_offset, peb_size - ubi_info.data_offset);
		}
	} else {
		fprintf(stderr, "PEB contained no valid UBI/UBIFS data\n");
	}
	peb_len = 0;
}


/* copy data to peb buffer
   return true if peb filled
*/
static int fill_peb(char *buf, struct range_data *rd)
{
	int len;

	len = MIN(rd->src_len, peb_size - peb_len );

	memcpy(peb + peb_len, buf + rd->src_offset , len);

	peb_len += len;
	rd->src_len -= len;
	rd->src_offset += len;

	if (peb_len == peb_size) {
		return 1;
	}

	return 0;
}

static int write_data(char *buf,int offset, int len)
{
        struct range_data rd;

        if (!active)
                return 0;

        rd = adjust_range(data_start, data_end, offset, len);

        while (rd.src_len ) {
		if (no_ubi) {
			/* only ubifs data write it directly no need to touch the data */
			write(cmd->io_out, buf + rd.src_offset, rd.src_len);
			break;
		}else{
			if (fill_peb(buf, &rd)) {
				extract_ubifs();
			}
		}
	}
        return 0;
}

static void usage(void)
{
        printf("      -u, --ubivol=NAME\tUBI volume device file to write rootfs to.\n");
	printf("      -U, --ubifs=NAME\tExtract ubifs image to file.\n");
}

static void ubi_opt( int argc, char **argv){
        int ch;

        while (1) {
                int option_index = 0;
                static struct option long_options[] = {
                        {"ubivol",  required_argument, 0, 'u'},
			{"ubidump",  required_argument, 0, 'U'},
                        {0, 0, 0, 0}
                };

                ch = getopt_long(argc, argv, "U:u:",
                                 long_options, &option_index);

                if (ch == -1)
                        break;

                switch (ch) {
                case 'u':
                        active = 1;
			snprintf(ubifs_device, 100,"/dev/%s", optarg);
                        break;
                case 'U':
			filename = optarg;
			active = 1;
                        dump = 1;
                        break;
                default:
                        ;
                }
        }
}

static struct ops my_ops = {
        .description = "UBIFS filesystem handler.",
        .init = init,
        .usage = usage,
        .opt = ubi_opt,
        .data = write_data,

};

CONSUMER(my_ops);

